#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Mar 12 20:17:29 2019

@author: marius
"""

import pyomo.environ as pyo
from math import floor

class ZEBModel():
    
    def __init__(self, M_const = 1000):
        """Create Abstract Pyomo model for ZEB
        """
        # Deterministic two-stage model
        self.abstractmodel = self.createTWOSTAGEMODEL()  
        
        self.M_const = M_const   
         
    def disco(self, n, r):
        '''Discounting factor'''
        return 1/((1+r)**n)
        
    def annui(self, n, r): 
        '''Annuity factor'''
        return r/(1-(1+r)**(-n))
        
    def capit(self, n, r): # Capitalization factor
        '''Capitalization factor'''
        return (1-(1+r)**(-n))/r
        
    def cost(self, Yn, cost, l, r): 
        '''For the two-stage model: calculating forced reinvestment costs'''
        Kn = pyo.floor(Yn/(l*1)) 
        n = Yn-l*Kn 
        Tn = Yn-n 
        return cost*(self.annui(l,r)*self.capit(n,r)*self.disco(Tn, r) \
                 + sum(self.disco(k*l,r) for k in range(0,Kn)))

        
    def npv_cost_Investments(self, m, st):
        '''Investment function for two-stage/deterministic, including o&m costs'''
        investments = 0
        if st==1:
            for i in m.I: 
                investments +=  (m.C_spe[i]*m.x[i]+ m.C_fxd[i]*m.a_i[i])
                '''Operations and maintenance costs included in investment costs:'''                
                #investments +=  (m.C_run[i]*m.C_spe_0[i]*m.x[i])*self.capit(m.YRN, m.R)*self.disco(1, m.R)
        else:
            investments = 0
        return investments
    
    def npv_cost_Operations(self, m, st): 
        '''Operational costs for  two-stage/deterministic
        Summation of yearly costs for all years in YRN'''
        techrun = 0
        runcosts = 0
        gridtariff = 0
        operations = 0
        if m.lastT == 4367:
            f = 2
        elif m.lastT == 23:
            f = 8736/24
        elif m.lastT == 287:
            f = 8736/288
        elif m.lastT == 671:
            '''multiplication factor will be 13 if reduced model'''
            f = 13
        elif m.lastT == 727:
            f = 12
        elif m.lastT == 8735:
            f = 1
            
        if st == 2:
            for i in m.I:
                techrun +=  m.C_run[i]*m.C_spe_0[i]*m.x[i]
            if m.A_ep: #Grid tariff model (includes VAT): Energy pricing
                ''' Nettleie: fastledd=8.61 EUR/mnd, energiledd = 0.05 EUR/kWh (inkluderer enova-avgift samnt moms)'''
                gridtariff = 12*m.C_fxd_ep + m.C_spe_ep*sum(f*m.y_imp[t] for t in m.T)
            elif m.A_ps == 1: #Grid tariff model(includes VAT): Power subscription
            
                gridtariff =  12*(m.C_fxd_ps*(1+m.Y_max)) + m.C_pty_ps*sum(f*m.y_pty[t] for t in m.T) + m.C_spe_ps*sum(f*m.y_imp[t] for t in m.T)
    
            VAT = 1.25 
            runcosts  = sum(m.f_fp[t]*m.C_wo + VAT*m.y_imp[t]*m.P_spot[t] - m.y_exp[t]*m.P_spot[t]*m.A_exp for t in m.T)
            operations = (f*runcosts + gridtariff + techrun)*self.capit(m.YRN, m.R)*self.disco(1, m.R)
        else:
            operations = 0
        return operations
    
    
    def createTWOSTAGEMODEL(self):
        m = pyo.AbstractModel()
        m.name = 'ZEB stochastic two-stage model'
    
    # SETS ##########################################################
        m.T  = pyo.Set(doc = 'Set of all hours, full model: 8736, reduced model: 672')
        m.M = pyo.Set(doc = 'Set of all months')
        m.I = pyo.Set(doc = 'Set of all technologies')
        m.ST = pyo.Set(initialize = [1, 2], doc='STAGE')    
    
    # PARAMETERS ##########################################################

        m.lastT = pyo.Param(within=m.T, doc="Last time step")

    #---Technology costs
        m.C_fxd_0 = pyo.Param(m.I,within=pyo.NonNegativeReals, default = 0, doc='Fixed investment cost for all techs, EUR in year t = 0')
        
        m.C_spe_0 = pyo.Param(m.I,within=pyo.NonNegativeReals, default = 0,doc='Investment costs dependent on installed capacity, EUR/kW (EUR/kWh) in t= 0')
        m.C_run = pyo.Param(m.I,within=pyo.NonNegativeReals,default = 0,doc='Yearly running cost of each tech, given from investment costs EUR/kW installed')
    
    #---Grid Tariff pricing
        #Energy pricing 
        m.A_ep = pyo.Param(within=pyo.Binary,default = 0,doc = 'Activation of energy pricing')
        m.C_fxd_ep = pyo.Param(within=pyo.NonNegativeReals,default = 0,doc='Fixed charge part of grid tariff for ep')
        m.C_spe_ep = pyo.Param(within=pyo.NonNegativeReals,default = 0,doc='Specific energy charge part of grid tariff for ep')
    
        #Power Subscription pricing
        m.A_ps = pyo.Param(within = pyo.Binary, default = 0,doc = 'Activation of power subscription pricing')  
        m.C_fxd_ps = pyo.Param(within=pyo.NonNegativeReals,default = 0, doc='Subscriptopn charge for pp')
        m.C_pty_ps = pyo.Param(within=pyo.NonNegativeReals,default = 0,doc='Penalty charge for pp')
        m.C_spe_ps = pyo.Param(within=pyo.NonNegativeReals,default = 0,doc='Energy charge charge for pp')
        m.Y_max = pyo.Param(within=pyo.NonNegativeReals,default = 0,doc = 'Subscription limit')
    
        #Peak Power Pricing
        m.A_pp = pyo.Param(within=pyo.Binary)
        m.C_pp = pyo.Param(within=pyo.NonNegativeReals)
        m.C_fxd_pp = pyo.Param(within=pyo.NonNegativeReals)                                   
        m.C_spe_pp = pyo.Param(within=pyo.NonNegativeReals)  
    
    #---#Reference System
    
        #CO2-Factors
        m.A_co2 = pyo.Param(within=pyo.Binary,doc = 'Activation of co2 crediting system')
        m.G_ref = pyo.Param(within=pyo.NonNegativeReals,doc='CO2 reference emissions')
        m.G_el_imp = pyo.Param(within=pyo.NonNegativeReals,doc='gCO2 eq. per kWh imported/exported')
        m.G_el_exp = pyo.Param(within=pyo.NonNegativeReals,doc='gCO2 eq. per kWh imported/exported')
        m.G_bf = pyo.Param(within=pyo.NonNegativeReals,doc='gCO2 eq. per kWh for technology i, i.e BB')
        m.G_wo = pyo.Param(within=pyo.NonNegativeReals,doc='gCO2 eq. per kWh for wood technology')
        m.G_tot = pyo.Param(within=pyo.NonNegativeReals,doc='gCO2 total for noZEB')
        
        #Primary Energy Factors
        m.A_pe = pyo.Param(within=pyo.Binary,doc = 'Activation of primary energy crediting system')
        m.PE_ref = pyo.Param(within=pyo.NonNegativeReals,doc='CO2 reference emissions')
        m.PE_imp = pyo.Param(within=pyo.NonNegativeReals,doc='PE per kWh imported electricity')
        m.PE_exp = pyo.Param(within=pyo.NonNegativeReals,doc='PE per kWh exported electricity')
        m.PE_bf = pyo.Param(within=pyo.NonNegativeReals,doc='PE per kWh for technology i, i.e BB')
        
    #---Technologies
        m.A_i = pyo.Param(m.I,  within=pyo.Binary,doc='Pre-activation of each tech')
        
        m.Eff = pyo.Param(m.I,within=pyo.NonNegativeReals, doc='Technology efficiency')
        m.Eff_ba_ch = pyo.Param(within=pyo.NonNegativeReals,doc='Battery charging efficiency')
        m.Eff_ba_dch = pyo.Param(initialize = 1, doc='Battery discharge efficiency')
        m.Beta_ba = pyo.Param(within=pyo.NonNegativeReals,doc='Charging/discharging rate')
        m.Beta_hs = pyo.Param(within=pyo.NonNegativeReals, doc='identical charging rate for heat storage')
        
        m.L = pyo.Param(m.I,  within=pyo.NonNegativeIntegers, doc='Lifetime of technology i')
        m.X_min = pyo.Param(m.I,  within=pyo.NonNegativeReals, doc='Max possible installed capacity of technology ')
        m.X_max = pyo.Param(m.I, within=pyo.NonNegativeReals, doc='Min possible installed capacity of technology ')
        
        m.Temp = pyo.Param(m.T, within=pyo.Reals, doc='Ambient temperature of certain hour')
        m.Y_pv = pyo.Param(m.T, within=pyo.NonNegativeReals,doc='Possible PV output at time t')
        m.COP_a2a = pyo.Param(m.T, within=pyo.NonNegativeReals, doc='Air-to-air heat pump performance at time t w/r SH')

        
        #BITES-params
        m.SS_cap = pyo.Param(within=pyo.NonNegativeReals, doc='Shallow storage capacity of house, fully determined by area, thus a parameter instead of a variable')
        m.DS_cap = pyo.Param(within=pyo.NonNegativeReals, doc='Deep storage capacity of house, ---=---')
        m.K_shallow = pyo.Param(doc='loss factor for BITES, shallow part')
        m.K_deep = pyo.Param(doc='loss factor for BITES, deep part')
        m.K_flow = pyo.Param(doc='flow factor for BITES, cross-node flow')
        
    #---Energy Demand    
        m.D_el = pyo.Param(m.T,  within=pyo.NonNegativeReals, doc='Hourly building electricity demand')
        m.D_sh = pyo.Param(m.T, within=pyo.NonNegativeReals, doc='Hourly building space heating demand')
        m.D_dhw = pyo.Param(m.T, within=pyo.NonNegativeReals, doc='Hourly building domestic hot water demand')
        
    #---Grid
        m.P_spot = pyo.Param(m.T,  within=pyo.NonNegativeReals, doc='Hourly price of imported electricity EUR/kWh including certificates')
        m.X_max_imp = pyo.Param(within=pyo.NonNegativeReals, doc='Maximum grid import')
        m.X_max_exp = pyo.Param( within=pyo.NonNegativeReals,doc='Maximum grid export')
        m.A_imp = pyo.Param( within=pyo.Binary, doc='Binary: Import is activated, 1/0')
        m.A_exp = pyo.Param( within=pyo.Binary,doc='Binary: Export is activated, 1/0')
        m.C_bf = pyo.Param(within=pyo.NonNegativeReals, doc='Constant price of biofuel')
        m.C_wo = pyo.Param(within=pyo.NonNegativeReals, doc='Constant price of wood')
        
    #---Control
        m.gamma = pyo.Param( within=pyo.NonNegativeReals, doc='=0 for strictly ZEB')
        m.R = pyo.Param(within=pyo.NonNegativeReals, doc='Chosen discount Rate')
        m.YRN = pyo.Param( within=pyo.NonNegativeIntegers, doc='Total years in modelling period') 
        
        def npv_inv_spe(m, i): 
            return self.cost(m.YRN, m.C_spe_0[i], m.L[i], m.R)
        m.C_spe = pyo.Param(m.I, rule = npv_inv_spe)
        
        def npv_inv_fxd(m, i): 
            return self.cost(m.YRN, m.C_fxd_0[i], m.L[i], m.R)
        m.C_fxd = pyo.Param(m.I, rule = npv_inv_fxd)

    #VARIABLES ##########################################################  
        
        # 1 STAGE : STRATEGIC VARIABLES
        m.x = pyo.Var(m.I,within = pyo.NonNegativeReals,                          doc='Optimal installed capacity (storage size), semi-continous, kW (kWh)')  
        m.a_i = pyo.Var(m.I, within = pyo.Binary,                                  doc='Activation binary decision for technology i, 1/0')
        
        
        #2 STAGE : OPERATIONAL VARIABLES
        m.q_hs = pyo.Var(m.T, domain = pyo.Reals,                                  doc='Keeping track of HS discharge')
        m.q_eb_dhw = pyo.Var(m.T,domain = pyo.NonNegativeReals,                        doc='Net heat supplied from (DHW) electric boiler at time t, kWh/h')
        m.q_po_sh = pyo.Var(m.T,domain = pyo.NonNegativeReals,                        doc='Net heat supplied from (SH) paneloven at time t, kWh/h')
        m.q_a2a_sh = pyo.Var(m.T,  domain= pyo.NonNegativeReals,                          doc='Net heat supplied from air-to-air heat pump at time t, kWh/h')
        m.q_fp_sh = pyo.Var(m.T,  domain=pyo.NonNegativeReals,                         doc='Net heat supplied from bio boiler at time t, kWh/h')
        m.f_fp = pyo.Var(m.T,  domain= pyo.NonNegativeReals,                          doc='Wood input to the fireplace at time t kWh/h')
        
        m.z_hs = pyo.Var(m.T,  domain = pyo.NonNegativeReals,                       doc='Content in heat storage at the end of time t, kWh')
        m.z_ss = pyo.Var(m.T,  domain = pyo.NonNegativeReals,                       doc='Content in shallow BITES at the end of time t, kWh')
        m.z_ds = pyo.Var(m.T,  domain = pyo.NonNegativeReals,                       doc='Content in deep BITES at the end of time t, kWh')
        m.z_ba = pyo.Var(m.T,  domain = pyo.NonNegativeReals,                       doc='Content of battery at the end of time t, kWh')
        
        #charging variables for BITES, shallow part
        #m.z_ss_ch = pyo.Var(m.T,  domain = pyo.NonNegativeReals,                       doc='Maximum charging rate, shallow BITES at the end of time t, kWh')
        #m.z_ss_dch = pyo.Var(m.T,  domain = pyo.NonNegativeReals,                      doc='Maximum discharging rate, shallow BITES at the end of time t, kWh')
        m.q_ss = pyo.Var(m.T,  domain = pyo.Reals,                      doc='Energy released from shallow BITES at the end of time t, kWh')
        m.Flow = pyo.Var(m.T,  domain = pyo.Reals,                                  doc='Cross-node flow of two-node BITES model at time t, kWh')
        
        
        m.y_imp = pyo.Var(m.T, domain = pyo.NonNegativeReals,                      doc='Electricity imported from grid at time t, kWh')
        m.y_exp = pyo.Var(m.T,  domain = pyo.NonNegativeReals,                      doc='Electricity exported to grid at time t, kWh')
        m.y_pv = pyo.Var(m.T, domain = pyo.NonNegativeReals,                       doc='PV production at time t, kWh/h')
        m.y_eb_dhw = pyo.Var(m.T, domain = pyo.NonNegativeReals,                       doc='Electricity drawn from (DHW) electric boiler at time t, kWh/h')
        m.y_po_sh = pyo.Var(m.T, domain = pyo.NonNegativeReals,                       doc='Electricity drawn for (SH) paneloven at time t, kWh/h')
        m.y_a2a_sh = pyo.Var(m.T,  domain=pyo.NonNegativeReals,                         doc='Electricity drawn for air-to-air heat pump at time t, kWh/h')
        m.y_ch = pyo.Var(m.T, domain = pyo.NonNegativeReals,                       doc='Amount of electricity to battery (charging) at time t, kWh/h')
        m.y_dch = pyo.Var(m.T, domain = pyo.NonNegativeReals,                      doc='Amount of electricity discharge from battery at time t, kWh/h')
        m.y_pty = pyo.Var(m.T, domain = pyo.Reals,                                 doc = 'Penalty volume')
        
        
        m.y_max = pyo.Var(m.M, domain = pyo.NonNegativeReals,                      doc='max power for every month')   
    
        m.a_ch = pyo.Var(m.T,  domain = pyo.Binary,                                 doc ='Charging activation inward time t, 1 = activated')
        m.a_dch = pyo.Var(m.T,  domain= pyo.Binary,                                 doc ='Discharging activation inward time t, 1=activated')
    
        
    # CONSTRAINTS ##########################################################
        
    # 1 STAGE : INVESTMENTS
        
        #---Activation and boundary constraints 
        def Tech_active(m, i, st):
            return m.x[i] <= m.a_i[i]*self.M_const
        m.Tech_active = pyo.Constraint(m.I, m.ST, rule = Tech_active) 
        
        def Tech_Min(m, i, st):
            return m.X_min[i]*m.a_i[i] <= m.x[i] 
        m.Tech_Min = pyo.Constraint(m.I, m.ST, rule= Tech_Min)
       
        def Tech_Max(m, i, st): 
            return m.x[i] <= m.X_max[i]*m.A_i[i]
        m.Tech_Max = pyo.Constraint(m.I, m.ST, rule= Tech_Max)

    #2 STAGE : OPERATIONS
         
        #---Balacing constraints

        def El_Balance(m, t, st): 
            return m.D_el[t] == m.y_imp[t] + m.y_pv[t] - m.y_exp[t] + m.y_dch[t] - m.y_ch[t] - m.y_a2a_sh[t] - m.y_po_sh[t] - m.y_eb_dhw[t]
        m.El_Balance = pyo.Constraint(m.T, m.ST, rule = El_Balance)

        
        def SH_balance(m, t, st):                   
            return m.D_sh[t] == m.q_ss[t] + m.q_po_sh[t] + m.q_a2a_sh[t] + m.q_fp_sh[t]
        m.SH_balance = pyo.Constraint(m.T, m.ST, rule = SH_balance)
          
        def DHW_balance(m, t, st):
            if t == 0:
                return m.D_dhw[t] + m.z_hs[t] == m.z_hs[m.lastT]*m.Eff['HS'] + m.q_eb_dhw[t]
            else:
                return m.D_dhw[t] + m.z_hs[t] == m.z_hs[t-1]*m.Eff['HS'] + m.q_eb_dhw[t]  
        m.DHW_balance = pyo.Constraint(m.T, m.ST, rule = DHW_balance)
            
        #---Capacity
        #split heat pumps in two here as well
        #point heat, no heat pumps
        def PO_SH_Restriction(m,t, st): 
            return m.q_po_sh[t] <= m.x['PO_SH']
        m.PO_SH_Restriction = pyo.Constraint(m.T, m.ST, rule = PO_SH_Restriction)
        
        def FP_SH_Restriction(m,t, st):
            i = t - floor(t/24)*24
            if i >= 16 and i <= 24:
                return m.q_fp_sh[t] <= m.x['FP_SH']
            else:
                return m.q_fp_sh[t] == 0
        m.FP_SH_Restriction = pyo.Constraint(m.T, m.ST, rule = FP_SH_Restriction)
        
        def A2A_SH_Restriction(m,t, st): 
            return m.q_a2a_sh[t] <= m.x['A2A_SH']
        m.A2A_SH_Restriction = pyo.Constraint(m.T, m.ST, rule = A2A_SH_Restriction)

        def A2A_SH_Restriction_1(m,t, st): 
            return m.q_a2a_sh[t] <= 0.4*m.D_sh[t]
        m.A2A_SH_Restriction_1 = pyo.Constraint(m.T, m.ST, rule = A2A_SH_Restriction_1)
        
        def EB_DHW_Restriction(m,t, st): 
            return m.q_eb_dhw[t] <= m.x['EB_DHW']
        m.EB_DHW_Restriction = pyo.Constraint(m.T, m.ST, rule = EB_DHW_Restriction)
              
        #---Grid equations    
        def Grid_Import(m,t, st): 
            return m.y_imp[t] <= m.X_max_imp
        m.Grid_Import = pyo.Constraint(m.T, m.ST,  rule=Grid_Import)
                
        def Grid_Export(m,t, st): 
            return m.y_exp[t] <= m.X_max_exp
        m.Grid_Export = pyo.Constraint(m.T, m.ST,  rule=Grid_Export)
                      
        #---Storage equations  

        def HS_Restriction(m, t, st): 
            return m.z_hs[t] <= m.x['HS']
        m.HS_Restriction = pyo.Constraint(m.T, m.ST, rule=HS_Restriction)

        def SS_Restriction(m, t, st): 
            return m.z_ss[t] <= m.x['SS']*m.A_i['SS'] 
        m.SS_Restriction = pyo.Constraint(m.T, m.ST, rule=SS_Restriction)
        
        def DS_Restriction(m, t, st): 
            return m.z_ss[t] <= m.x['DS']*m.A_i['DS'] 
        m.DS_Restriction = pyo.Constraint(m.T, m.ST, rule=DS_Restriction)
        
        def SS_charge_active(m,t, st): 
            return m.q_ss[t] <= m.z_ss[t]
        m.SS_charge_active = pyo.Constraint(m.T, m.ST, rule=SS_charge_active)
    
        def SS_discharge_active(m,t, st): 
            return - m.z_ss[t] <= m.q_ss[t]
        m.SS_discharge_active = pyo.Constraint(m.T, m.ST, rule=SS_discharge_active)
       
        def HS_charge_active(m,t, st): 
            return m.q_hs[t] <= m.z_hs[t]
        m.HS_charge_active = pyo.Constraint(m.T, m.ST, rule=HS_charge_active)
        
        def HS_discharge_active(m,t, st): 
            return - m.z_hs[t] <= m.q_hs[t]
        m.HS_discharge_active = pyo.Constraint(m.T, m.ST, rule=HS_discharge_active)
        '''
        def SS_charge_rate_max(m, t, st):
            if m.Temp[t] >= 15:
                return m.q_ss[t] >= -m.x['SS']
            elif m.Temp[t] < 15 and m.Temp[t] > -15:
                return m.q_ss[t] <= m.x['SS'] * (1 - (15 - m.Temp[t])/30)
            else:
                return m.q_ss[t] >= 0
        m.SS_charge_rate_max = pyo.Constraint(m.T, m.ST, rule = SS_charge_rate_max)
        
        def SS_discharge_rate_min(m, t, st):
            if m.Temp[t] >= 15:
                return m.q_ss[t] <= 0
            elif m.Temp[t] < 15 and m.Temp[t] > -15:
                return m.q_ss[t] >= -(m.x['SS'] * ((15 - m.Temp[t])/30))
            else:
                return m.q_ss[t] <= m.x['SS']
        m.SS_discharge_rate_min = pyo.Constraint(m.T, m.ST, rule = SS_discharge_rate_min)
        '''
             
        def HS_Balance_ch(m,t,st):
            if t == 0:
                return m.q_hs[t] == m.z_hs[m.lastT] - m.z_hs[t]
            else:
                return m.q_hs[t] == m.z_hs[t-1] - m.z_hs[t]    
        m.HS_Balance_ch = pyo.Constraint(m.T, m.ST, rule = HS_Balance_ch)
        
        '''  
        def SS_balance_ch(m, t, st):
            if t == 0:
                return m.q_ss[t] == m.z_ss[m.lastT] - m.z_ss[t]
            else:
                return m.q_ss[t] == m.z_ss[t-1] - m.z_ss[t]
        m.SS_balance_ch = pyo.Constraint(m.T, m.ST, rule = SS_balance_ch)
        ''' 
        def Flow_Constraint(m,t,st):
            return m.Flow[t] == m.K_flow*((m.z_ss[t]/m.SS_cap) - (m.z_ds[t]/m.DS_cap))
        m.Flow_Constraint = pyo.Constraint(m.T, m.ST, rule = Flow_Constraint)
        
        def SS_balance_ch(m, t, st):
            if t == 0:
                return m.q_ss[t] == m.z_ss[m.lastT] - m.z_ss[t] - m.Flow[t] - m.z_ss[t]*(1 - m.K_shallow)
            else:
                return m.q_ss[t] == m.z_ss[t-1] - m.z_ss[t] - m.Flow[t] - m.z_ss[t]*(1 - m.K_shallow)
        m.SS_balance_ch = pyo.Constraint(m.T, m.ST, rule = SS_balance_ch)
        
        def DS_balance_ch(m, t, st):
            if t == 0:
                return m.z_ds[t] == m.z_ds[m.lastT] + m.Flow[t] - m.z_ds[t]*(1 - m.K_deep)
            else:
                return m.z_ds[t] == m.z_ds[t-1] + m.Flow[t] - m.z_ds[t]*(1 - m.K_deep)
        m.DS_balance_ch = pyo.Constraint(m.T, m.ST, rule = DS_balance_ch)
        '''
        def SS_discharge_rate_min(m,t, st): 
               return -m.x['SS'] <= m.q_ss[t] 
        m.SS_discharge_rate_min = pyo.Constraint(m.T, m.ST, rule=SS_discharge_rate_min)
        
        def SS_discharge_rate_max(m,t, s): 
            return m.q_ss[t] <= m.x['SS'] 
        m.SS_discharge_rate_max = pyo.Constraint(m.T, m.ST, rule=SS_discharge_rate_max)
        '''
           
        def HS_discharge_rate_min(m,t, st): 
               return -m.x['HS']*m.Beta_hs <= m.q_hs[t]
        m.HS_discharge_rate_min = pyo.Constraint(m.T, m.ST, rule=HS_discharge_rate_min)
                        
        def HS_discharge_rate_max(m,t, s): 
            return m.q_hs[t] <= m.x['HS']*m.Beta_hs
        m.HS_discharge_rate_max = pyo.Constraint(m.T, m.ST, rule=HS_discharge_rate_max)
     
        def BA_Restriction(m,t, st): 
            return m.z_ba[t] <= m.x['BA'] 
        m.BA_restriction = pyo.Constraint(m.T, m.ST, rule=BA_Restriction)
           
        def BA_Balance(m,t, st):
            if t == 0:
                return m.z_ba[t] == m.z_ba[m.lastT] - m.y_dch[t]*(1/m.Eff_ba_dch) + m.y_ch[t]*m.Eff_ba_ch
            else:
                return m.z_ba[t] == m.z_ba[t-1] - m.y_dch[t]*(1/m.Eff_ba_dch) + m.y_ch[t]*m.Eff_ba_ch
        m.BA_Balance = pyo.Constraint(m.T, m.ST, rule=BA_Balance)
               
        def BA_Charge_Balance(m, t, st): 
            if t == 0:
                return m.y_ch[t] <= (m.x['BA'] - m.z_ba[m.lastT])*m.A_i['BA']*(1/m.Eff_ba_ch)
            else:
                return m.y_ch[t] <= (m.x['BA'] - m.z_ba[t-1])*m.A_i['BA']*(1/m.Eff_ba_ch)
        m.BA_Charge_Balance = pyo.Constraint(m.T, m.ST, rule=BA_Charge_Balance)
               
        def BA_Discharge_Balance(m,t, st): 
            if t == 0:
                return m.y_dch[t] <= m.z_ba[m.lastT]*m.A_i['BA']*m.Eff_ba_dch
            else:
                return m.y_dch[t] <= m.z_ba[t-1]*m.A_i['BA']*m.Eff_ba_dch
        m.BA_Discharge_Balance = pyo.Constraint(m.T, m.ST, rule=BA_Discharge_Balance)
    
        def BA_charge_active(m,t, st): 
            return m.y_ch[t] <= m.X_max_imp*m.a_ch[t]
        m.BA_charge_active = pyo.Constraint(m.T, m.ST, rule=BA_charge_active)
    
        def BA_discharge_active(m,t, st): 
            return m.y_dch[t] <= m.X_max_imp*m.a_dch[t]
        m.BA_discharge_active = pyo.Constraint(m.T, m.ST, rule=BA_discharge_active)
    
        def Battery_Balance(m,t, st): 
            return m.a_ch[t] + m.a_dch[t] <= 1
        m.Battery_Balance = pyo.Constraint(m.T, m.ST, rule=Battery_Balance)
    
        def BA_charge_rate(m,t, st): 
            return m.y_ch[t] <= m.x['BA']*m.Beta_ba
        m.BA_charge_rate = pyo.Constraint(m.T, m.ST, rule=BA_charge_rate)
    
        def BA_discharge_rate(m,t, st): 
            return m.y_dch[t] <= m.x['BA']*m.Beta_ba
        m.BA_discharge_rate = pyo.Constraint(m.T, m.ST, rule=BA_discharge_rate)
       
        #---Production constraints for generating technologies
   
        def PV_Balance(m,t, st): 
            return m.y_pv[t] == m.x['PV']*m.Y_pv[t] 
        m.PV_Balance = pyo.Constraint(m.T, m.ST, rule=PV_Balance)
    
        def A2A_SH_Balance(m,t, st): 
            return m.q_a2a_sh[t] == m.y_a2a_sh[t]*m.COP_a2a[t] 
        m.A2A_SH_Balance = pyo.Constraint(m.T, m.ST, rule = A2A_SH_Balance)
        
        def PO_SH_Balance(m,t, st): 
            return m.q_po_sh[t] == m.y_po_sh[t]*m.Eff['PO_SH'] 
        m.PO_SH_Balance = pyo.Constraint(m.T, m.ST, rule = PO_SH_Balance)
        
        def FP_SH_Balance(m,t, st): 
            return m.q_fp_sh[t] == m.f_fp[t]*m.Eff['FP_SH']*m.A_i['FP_SH'] 
        m.FP_SH_Balance = pyo.Constraint(m.T, m.ST, rule = FP_SH_Balance)
        
        def EB_DHW_Balance(m,t, st): 
            return m.q_eb_dhw[t] == m.y_eb_dhw[t]*m.Eff['EB_DHW'] 
        m.EB_DHW_Balance = pyo.Constraint(m.T, m.ST, rule = EB_DHW_Balance)
        
        
        #---Zero emission/energy constraints    
        def ZE_Balance(m): 
            if m.A_co2==1:
                print('ACTIVE ZEB-carbon RESTRICTION')
                if m.lastT == 8735:
                    return sum(m.y_imp[t]*m.G_el_imp + m.f_fp[t]*m.G_wo for t in m.T) <=  sum(m.y_exp[t]*m.G_el_exp for t in m.T)
                if m.lastT == 4367:
                    return 2*sum(m.y_imp[t]*m.G_el - m.y_exp[t]*m.G_el + m.bf[t]*m.G_bf for t in m.T) <= m.G_ref*m.gamma
                elif m.lastT == 727:
                    return 12*sum(m.y_imp[t]*m.G_el - m.y_exp[t]*m.G_el + m.bf[t]*m.G_bf for t in m.T) <= m.G_ref*m.gamma
                elif m.lastT == 671:
                    return sum(m.y_imp[t]*m.G_el_imp + m.f_fp[t]*m.G_wo for t in m.T) <= sum(m.y_exp[t]*m.G_el_exp for t in m.T)
                elif m.lastT == 287:
                    return (8736/288)*sum(m.y_imp[t]*m.G_el - m.y_exp[t]*m.G_el + m.bf[t]*m.G_bf for t in m.T) <= m.G_ref*m.gamma
                elif m.lastT == 23:
                    return (8736/24)*sum(m.y_imp[t]*m.G_el - m.y_exp[t]*m.G_el + m.bf[t]*m.G_bf for t in m.T) <= m.G_ref*m.gamma
            elif m.A_pe == 1:
                print('ACTIVE ZEB-pef RESTRICTION')
                if m.lastT == 8735:
                    return sum(m.y_imp[t]*m.PE_imp - m.y_exp[t]*m.PE_exp + m.bf[t]*m.PE_bf for t in m.T) <= m.PE_ref*m.gamma 
                elif m.lastT == 671:    
                    return 13*sum(m.y_imp[t]*m.PE_imp - m.y_exp[t]*m.PE_exp + m.bf[t]*m.PE_bf for t in m.T) <= m.PE_ref*m.gamma  
            else:
                print('NO ZEB RESTRICTION')
                return pyo.Constraint.Skip
        m.ZE_Balance = pyo.Constraint(rule=ZE_Balance) 
                 
    #Subscription power pricing Constraint
        def pty_volume(m, t): #counting all power within one hour exceeding max limit
            if m.A_ps == 1:
                return m.y_imp[t] - m.Y_max <= m.y_pty[t]
            else:
                return m.y_pty[t] == 0 
        m.pty_volume = pyo.Constraint(m.T, rule = pty_volume)
            
        def pty_volume2(m, t): 
                return 0 <= m.y_pty[t]
        m.pty_volume2 = pyo.Constraint(m.T, rule = pty_volume2)
    
    #   OBJECTIVE FUNCTION  #################################################

        def cost_Investments_rule(m, st):
            expr = self.npv_cost_Investments(m, st)
            return expr
        m.cost_Investments = pyo.Expression(m.ST, rule = cost_Investments_rule)

        def cost_Operation_rule(m, st):
            expr = self.npv_cost_Operations(m, st)
            return expr
        m.cost_Operations=  pyo.Expression(m.ST, rule = cost_Operation_rule)

        
        def objective_TotalCost(m): 
            expr = pyo.summation(m.cost_Investments) + pyo.summation(m.cost_Operations)
            return expr
        m.objective_TotalCost = pyo.Objective(rule = objective_TotalCost, sense = pyo.minimize)
           
        return m
    
    def createConcreteModeltwoStage(self, data):
            '''Function creating instance from input data'''
            #data = self.writeStochasticProblem(data)
            concretemodel = self.abstractmodel.create_instance(data = {'mymodel':data}, namespace = 'mymodel')
            return concretemodel